#include	"Command.h"
#include	"Interface.h"
#include	"Controller.h"
#include	"Board.h"
#include	"Engine.h"
#include	"Lawyer.h"
#include	"Evaluator.h"

#include	<iostream>
#include	<strstream>
#include	<cstdio>
#include	<cstdlib>
#include	<string>

/*
 * Command.cpp (c) Noah Roberts 2003-03-19
 */

using namespace std;

Command::Command(Interface *i, Board *b, Engine *e, Lawyer *l, Controller *c)
{
  iface   = i;
  board   = b;
  engine  = e;
  lawyer  = l;
  control = c;
}

// Performs command and returns true if a valid command string has been supplied.
bool Command::isCommand(string commandText)
{
  string 	cmd;
  vector<string> 	args;

  parseCommand(commandText, cmd, args);

  // The commands...
  if (cmd == "quit") exit(0);

  // 3 special commands for debugging - not documented.
  else if (cmd == "vomit") vomit();
  else if (cmd == "go") { Options::defaultOptions()->setValue("computerColor", board->sideToMove() == RED ? "r":"b"); }
  else if (cmd == "undo") undo();
  else if (cmd == "remove") retract();
  else if (cmd == "?") { if (engine->thinking()) engine->endSearch(); }
  else if (cmd == "new") _new();
  else if (cmd == "setboard") setBoard(commandText, args);
  else if (cmd == "t" || cmd == "time")
    {
      // Display the time left to make a move...
      (*iface) << "Time left is: "
               << (int)Timer::timerForColor(board->sideToMove())->timeLeft()
               << " second to make "
               << (int)Timer::timerForColor(board->sideToMove())->movesLeft()
               << " moves.\n";
    }
  else
    {
      return false; // not a command...
    }
  return true; // If we haven't returned then we did something...
}

/*
 * Private members
 */

void Command::parseCommand(string cmdText, string &cmd, vector<string> &args)
{
  istrstream	commandStream(cmdText.c_str(), cmdText.size());
  string	tmp;

  if (cmdText.size() < 1) return; // No arguments supplied.

  commandStream >> cmd;

  commandStream >> tmp;
  while (!commandStream.eof())
    {
      args.push_back(tmp);
      commandStream >> tmp;
    }
  switch (cmdText.at(cmdText.size() - 1)) // Sometimes there is a space at the end of the line...
    {
    case ' ':
    case '\t':
      cerr << "space\n";
      break;
    default:
      args.push_back(tmp); // Other times it is not a space and we probably want that information.
    }
}

void Command::vomit() // Vomit up some useful information for debugging.
{
  vector<Move> _moveHistory = board->history();
  cerr << "Starting Position: " << board->startingPosition() << "\n" // Positions...
       << " Current Position: " << board->getPosition() << "\n";
  vector<Move>::iterator it = _moveHistory.begin();
  for (int i = 1; it != _moveHistory.end(); i++, it++) // Move history...
    {
      if (i%2)
        fprintf(stderr, "%3d. ", i);
      if (i == 1)
        if (board->startingPosition().at(board->startingPosition().size() -1) == 'b')
          {
            cerr << "...  ";
            i ++;
          }
      cerr << (*it);
      if (i%2) cerr << " ";
      else cerr << endl;
    }
}

void Command::_new()
{
  engine->endSearch();
  (void)engine->getMove();
  board->resetBoard();
  iface->printBoard(*board);
}

void Command::setBoard(string cmdText, vector<string> args) // set up board based on supplied fen.
{
  if (args.size() < 2)
    {
      cerr << args.size() << endl;
      (*iface) << "Error (invalid format): " << cmdText << "\n";
    }
  else
    {
      engine->endSearch();
      (void)engine->getMove();
      board->setPosition(args[0] + " " + args[1]);
      iface->printBoard(*board);
    }
}

void Command::tellog(vector<string> args) // output arguments to cerr
{
  string msg;

  for (vector<string>::iterator it = args.begin(); it != args.end(); it++)
    {
      msg += (*it) + " ";
    }
  cerr << msg << endl;
}

void Command::showHashes() // Show zoberist keys...
{
  cerr << "Primary  : " << board->primaryHash() << " -> " << (board->primaryHash() & 262143) << endl;
  cerr << "Secondary: " << board->secondaryHash() << endl;
}

void Command::undo()
{
  if (board->history().empty() || Options::defaultOptions()->getValue("computerColor") != "n")
    (*iface) << "Error (command not legal now): undo\n";
  else
    {
      board->unmakeMove();
      iface->printBoard(*board);
    }
}

void Command::retract()
{
  if (board->history().size() > 1)
    {
      board->unmakeMove();
      board->unmakeMove();
      iface->printBoard(*board);
    }
  else
    (*iface) << "Error (command not legal now): remove\n";
}
